import os
import sys
import pandas as pd
import numpy as np
from pathlib import Path

p=str(Path(__file__).parent.parent)
'''Read BLUE Vitality'''

def chosen_DAP(f,y):
    '''This function returns the DAP chosen for regression for a valid combination of field and year'''
    dap={'2019':{'M':52,'V':36,'S':48},
         '2020':{'M':47,'V':51,'S':47},
         '2021':{'M':54,'V':58,'S':30}}

    return dap[y][f]


def read_normalized_BLUE_vitality(years=[],  chosen_dap=True):
    '''This function reads BLUE vitality measurements for the specified years into a dictionary of dictionaries,
    the vitality dictionary looks like {'yyyy':{'f1':df, 'f2':df,'f3':df},'yyyy':{'f1':df, 'f2':df,'f3':df}}.
    If chosen_dap is True the dataframes will only contain the column of that DAP'''
    folder=os.path.join(p, 'Extracted vitality/')

    normalized_vitality={}
    if len(years)==0:
        #read all years
        years_list=['2019', '2020','2021']
    else:
        years_list=years

    for y in years_list:
        normalized_vitality[y]={}
        for field in ['M','S','V', 'AVG', 'MS', 'MV', 'SV']:
            blues_list = [filename for filename in os.listdir(folder) if
                          (filename.split('_')[-2]==field and y in filename)]
            blue_df=pd.read_csv(os.path.join(folder, blues_list[0]))
            ordered_columns=['Variety', 'Batch']
            ordered_DAPS=blue_df.columns[2:].to_list()
            ordered_DAPS.sort()
            if field in ['M','S','V']:
                if chosen_dap:
                    good_dap=str(chosen_DAP(field,y))
                    ordered_columns.append(good_dap)
                else:
                    ordered_columns.extend(ordered_DAPS)
            else:
                ordered_columns.append('Avg')
            normalized_vitality[y][field]=blue_df[ordered_columns]

    return normalized_vitality


def read_train_test_data(trainon, teston, intervals, var_spec=False):
    num_intervals=intervals
    '''This function returns:
    --train_on_y: string containing the year we are training on, 
    --train_infof: string containing the field information we are training on, 
    --test_on_y: string containing the year we are testing on, 
    --test_on_f: string containing the field information we are testing on,
    --Y: dataframe containg labels and vigor measurements for training, 
    --X: dataframe containg labels and tuber measurements for training, 
    --Yt: dataframe containg labels and vigor measurements for testing, 
    --Xt: dataframe containg labels and tuber measurements for testing, 
    --num_intervals: number of intervals for the DG procedure, might differ from the input parameter inteervals.
    The input parameters to be specified are:
    -- trainon: string of the format f_yyyy where multiple fields can be concatenated, specifies training set,
    -- teston: string of the format f_yyyy where multiple fields can be concatenated, specifies testing set,
    -- intervals: integer, maximum number of intervals for the construction of the V matrix,
    -- var_spec: boolean, if false the model trained is variety agnostic, if true the test set choice is made in order
       to train variety specific models.
    '''
    switcht=False
    avgt=False
    switchtr=False
    avgtr=False
    if len(trainon.split('_'))>2:
        #we are training on two fields at the same time
        train_on_y1= trainon.split('_')[1]
        train_on_f1=trainon.split('_')[0]
        train_on_y2= trainon.split('_')[3]
        train_on_f2=trainon.split('_')[2]
        train_infof=f'{train_on_f1}{train_on_f2}'
        train_on_y=train_on_y2
        switchtr=True
    elif trainon.split('_')[0]=='AVG':
        avgtr=True
        train_infof='AVG'
        train_infoy=trainon.split('_')[1]
        train_on_y=train_infoy

    else:
        train_on_y=trainon.split('_')[1]
        train_on_f=trainon.split('_')[0]
        train_infof=f'{train_on_f}'

    if len(teston.split('_'))>2:
        #we are training on two fields at the same time
        test_on_f1= teston.split('_')[0]
        test_on_y2= teston.split('_')[3]
        test_on_f2= teston.split('_')[2]
        test_on_f=f'{test_on_f1}{test_on_f2}'
        test_on_y=test_on_y2
        switcht=True
    elif teston.split('_')[0]=='AVG':
        avgt=True
        test_on_f='AVG'
        test_infoy=teston.split('_')[1]
        test_on_y=test_infoy

    else:
        test_on_y=teston.split('_')[1]
        test_on_f=teston.split('_')[0]
        test_on_f=f'{test_on_f}'
        test_on_y=test_on_y

    Y1 = read_normalized_BLUE_vitality([train_on_y], True)

    if switchtr:
        # we are training on two fields
        Y=Y1[train_on_y][f'{train_on_f1}{train_on_f2}']
    elif avgtr:
        # we are training on all fields
        Y= Y1[train_on_y]['AVG']
    else:
        #we are training on a single year
        Y = Y1[train_on_y][train_on_f]


    X_all  = read_all_tuber_data(list(set([train_on_y, test_on_y])))

    X = X_all[train_on_y]
    X = X.groupby(by=['Variety', 'Batch'], as_index=False).mean()
    xcols = X.columns
    X['Batch'] = X['Batch'].astype(str)
    Y['Batch'] = Y['Batch'].astype(str)
    train_merged = pd.merge(Y, X, how='inner', on=['Variety', 'Batch'])
    if train_on_y=='2020':
        #drop outlier batch
        train_merged.drop(train_merged.loc[train_merged['Batch']=='26'].index, inplace=True)
    train_merged=train_merged.dropna()  # get rid of rows for which the X matrix is partially empty
    train_merged.index = np.arange(len(train_merged)) # avoid indexing problems by resetting index
    ########################################################
    # Read testing data
    ########################################################
    Xt = X_all[test_on_y]
    Xt = Xt.groupby(by=['Variety', 'Batch'], as_index=False).mean()

    Y1 = read_normalized_BLUE_vitality([test_on_y], True)

    if switcht:
        # we are testing on two fields
        Yt=Y1[test_on_y][f'{test_on_f1}{test_on_f2}']
    elif avgt:
        # we are testing on all fields
        Yt= Y1[test_on_y]['AVG']

    else:
        #we are testing on a single field
        Yt = Y1[test_on_y][test_on_f]

    # Yt = Yt.sort_values(by=Yt.columns[-1])
    xtcols = Xt.columns
    Xt['Batch'] = Xt['Batch'].astype(str)
    Yt['Batch'] = Yt['Batch'].astype(str)
    test_merged = pd.merge(Yt, Xt, how='inner', on=['Variety', 'Batch'])
    if test_on_y=='2020':
        # drop outlier batch
        test_merged.drop(test_merged.loc[test_merged['Batch']=='26'].index, inplace=True)
    test_merged=test_merged.dropna()  # get rid of rows for which the X matrix is partially empty
    test_merged.index = np.arange(len(test_merged))

    if train_on_y==test_on_y:
        print('Special test splitting')
        #then the X matrix will be the same in train and test ie we have to separate it already, it is the only case in
        # which we make a choice of the test set
        datasizetst=len(test_merged)
        test_size=int(0.3*datasizetst)
        if var_spec:
            test_index, num_intervals=choose_variety_proportional_test_varspec(test_size, test_merged, intervals)
            train_index=[x for x in train_merged['Batch'].values if x not in test_index]
        else:
            test_index, num_intervals=choose_variety_proportional_test(test_size, test_merged, intervals)
            train_index=[x for x in train_merged['Batch'].values if x not in test_index]

        # check that train index and test index do not share batches
        train_set = set(train_index)
        test_set = set(test_index)
        if train_set.intersection(test_set):
            print('  INTERSECTION NOT EMPTY   ')
        else:
            test_index_data = test_merged[test_merged['Batch'].isin(test_index)]
            indextest = list(test_index_data.index)
            train_index_data = train_merged[train_merged['Batch'].isin(train_index)]
            indextrain = list(train_index_data.index)
            if set(indextrain).intersection(set(indextest)):
                print('  INTERSECTION NOT EMPTY   AT INDEX LEVEl')
                sys.exit()
            else:
                train_index=indextrain
                test_index=indextest

    else:
        test_index=test_merged.index
        train_index=train_merged.index

    Y = train_merged.iloc[train_index, :3]
    X = train_merged.loc[train_index,xcols]

    Yt = test_merged.iloc[test_index, :3]
    Xt = test_merged.loc[test_index, xtcols]

    return train_on_y, train_infof, test_on_y, test_on_f, Y, X, Yt, Xt, num_intervals


def choose_variety_proportional_test(test_size, testdata, intervals):
    '''Separate the dataset of a single year in training and test set. This function is called only for the full field
    model, meaning the full dataset contains around 160-180 samples depending on the year. The choice is random, but a
    fair representation of all varieties in test and training is ensured.'''
    var_tst_prop=test_size//6

    test_index=[]
    for variety in testdata.Variety.unique():
        '''In this loop we make a random choice per variety'''
        var_df=testdata[testdata['Variety']==variety]
        indexvar = list(var_df.index)
        this_var_test = np.random.choice(indexvar,  var_tst_prop, replace=False)
        this_var_test = var_df.loc[this_var_test, 'Batch']
        test_index.extend(this_var_test)

    how_many_per_var=testdata[testdata['Batch'].isin(test_index)]['Variety'].value_counts()
    avail=[x for x in testdata['Batch'].values if x not in test_index]
    if not np.all(how_many_per_var>2):
        '''If there are fewer than 2 samples per variety in the test dataset we add some samples of the underrepresented 
        varieties'''
        for v in how_many_per_var[how_many_per_var<3].index:
            train_avail = testdata[testdata["Batch"].isin(avail)]
            vardata = train_avail[train_avail['Variety'] == v]
            addendum=np.random.choice(list(vardata.index),1)
            addendum = vardata.loc[addendum, 'Batch']
            test_index.extend(addendum)

    return test_index, intervals


def choose_variety_proportional_test_varspec(test_size, testdata, intervals):
    '''Separate the dataset of a single year in training and test set. This function is called only for the variety
    specific model, meaning the full dataset contains around 26-30 samples depending on the year.
     The choice is conditional on the viable number of intervals for the DG approach. Namely, we need to ensure that at
     least two datapoints are available per interval. We decrease the number of intervals until that is the case.
     Then we choose test points randomly in each interval avoiding the extremities of the interval.'''
    var_tst_prop=test_size//6
    test_index=[]
    intervals_dic={}
    for var in  testdata.Variety.unique():
        vardata = testdata[testdata['Variety'] == var]
        miny = np.min(vardata.iloc[:, 2])
        maxy = np.max(vardata.iloc[:, 2])
        biggertwo = False
        allfull = False
        intervals_while = intervals
        while not (biggertwo and allfull):
            delimiters = np.linspace(miny, maxy, intervals_while, endpoint=False)
            digiY = np.digitize(vardata.iloc[:, 2].to_numpy(), delimiters)
            ### check that there is at least two points in every interval
            vals, counts = np.unique(digiY, return_counts=True)
            biggertwo = np.all(counts > 2)
            if len(vals) == intervals_while:
                allfull = np.all(vals == range(1, intervals_while + 1))
            else:
                allfull = False
            intervals_while = intervals_while - 1

        intervals = intervals_while + 1

        intervals_dic[var]=intervals
        proportion_in_interval =int(np.ceil(var_tst_prop /intervals))
        for interval_num in range(1, intervals+1):

            data_this_int=vardata[digiY==interval_num]
            index_min=data_this_int[[data_this_int.columns[2]]].idxmin().item()
            index_max=data_this_int[[data_this_int.columns[2]]].idxmax().item()

            indexvar = list(vardata.index)
            indexvar.remove(index_min)
            indexvar.remove(index_max)

            this_var_test=np.random.choice(indexvar, min(len(indexvar),proportion_in_interval), replace=False)
            test_index.extend(this_var_test)
    return test_index, intervals_dic

def read_all_tuber_data(years):
    '''Read the dataframe containing all tuber data for the specified years'''
    folder='Tuber data'
    pathparent = os.path.abspath(__file__ + "/../../")#str(Path(__file__).parent.parent)
    all_t={}
    for y in years:
        filename=os.path.join(pathparent, f'{folder}/{y}_all_tuber.csv')
        print(filename)
        all_t[y]= pd.read_csv(filename , index_col=False)

    return all_t

