import numpy as np


def normalized_residual(true, pred):
    '''returns the normalized residuals between the two input vectors, true and pred. The residual is normalized by the
    2 norm of the true vector, so the order of the input parameters matters.'''

    true=true.reshape((-1,1))
    pred=pred.reshape((-1,1))
    return (np.linalg.norm(true-pred))/np.linalg.norm(true)


def unit_norm_zero_mean(arr):
    '''shifts and scales the input array so that the result will have mean zero and norm one'''
    unarr=arr-np.mean(arr)
    unarr=unarr/np.linalg.norm(unarr)
    if np.linalg.norm(arr)==0:
        print('Zero prediction')
        unarr=arr
    return unarr


def squared_res(arr1, arr2):
    '''returns the squared normalized residual between the two input invectors, similar to residual, the order of the
    input vectors matters, the first input vector is supposed to be the true measurement.'''
    num=np.linalg.norm(arr1-arr2)**2
    den=np.linalg.norm(arr1)**2
    return num/den


def learn_coefficient_matrix(V, Xh):
    '''Computes the matrix of coefficients A'''
    A=np.linalg.solve(V,Xh)

    return A


def v1k(delk, delkmin1, yi):
    '''First basis function for the DG scheme'''
    vi=(delk-yi)/(delk-delkmin1)
    return vi


def v2k(delk, delkmin1, yi):
    '''Second basis function of the DG scheme'''
    vi= (yi-delkmin1)/(delk-delkmin1)
    return vi


def V_DG(y , pieces):
    '''Computes the matrix V in the DG scheme. If the given number of segments yields a singular matrix V the function
    returns the couple (0,0), else it returns the matrices V (basis matrix), and W (matrix of weights).'''
    V = np.zeros((len(y), pieces*2))
    ymin=np.min(y)
    ymax=np.max(y)
    delimiters=np.linspace(ymin, ymax, pieces, endpoint=False)
    digiY=np.digitize(y, delimiters)

    ### check that there are at least two points in every interval
    vals, counts=np.unique(digiY, return_counts=True)
    biggertwo=np.all(counts>1)
    if len(vals)==pieces:
        allfull=np.all(vals==range(1,pieces+1))
    else:
        allfull=False

    if biggertwo and allfull:
        j=0
        for yi in digiY:
            try:
                delk=delimiters[yi]
            except:
                 delk=ymax
            delkmin1=delimiters[yi-1]

            v1=v1k(delk, delkmin1, y[j])
            v2=v2k(delk, delkmin1, y[j])
            V[j,( yi-1)*2]=v1
            V[j,(( yi-1)*2) +1] = v2

            j+=1

        W= np.zeros(len(y))
        jj=0
        for interval in range(1, pieces+1):
            nodes_in_interval=y[digiY==interval]
            lxdel=delimiters[interval-1]
            try:
                rxdel=delimiters[interval]
            except:
                rxdel=ymax
            iix=range(len(nodes_in_interval))
            w0=exterior_weight(nodes_in_interval[1], nodes_in_interval[0], lxdel)
            W[jj]=w0
            jj+=1
            for ii in iix[1:-1]:
                wj=interior_weight(nodes_in_interval[ii-1], nodes_in_interval[ii], nodes_in_interval[ii+1])
                W[jj]=wj
                jj+=1
            wlast=exterior_weight(nodes_in_interval[-2], nodes_in_interval[-1], rxdel)
            W[jj]=wlast
            jj+=1

        W= np.diag(W)
        return V, W
    else:

        return (0,0)


def interior_weight(lxpt, thispt, rxpt):
    '''Computes weight for an interior data point'''
    return (1/2)*(np.abs(lxpt-thispt)+np.abs(thispt-rxpt))


def exterior_weight(nextpt,  thispoint, delimiter):
    '''Computes weight for an exterior data point'''
    return (1/2)*(np.abs(nextpt-thispoint))+np.abs(thispoint-delimiter)



def V(yv, param):
    '''Calls V_DG where param is the number of intervals for the DG scheme'''
    return V_DG(yv, param)


def degree_spec_regression(V, W, y, X, yt, Xt, returnA=False):
    '''This function performs regression for a given V and W, different V and W can be created from the same data for
    different numbers of intervals. The function return the performance metrics for the regression and the prediction
    of y_test, optionally it can return the learnt coefficients A and a_0. '''
    VtWV = np.matmul(np.matmul(np.transpose(V ), W), V )
    VtW = np.matmul(np.transpose(V ), W)

    a_0 = learn_coefficient_matrix(VtWV, np.matmul(VtW, y))
    A = learn_coefficient_matrix(VtWV, np.matmul(VtW, X))

    # ## Predicting
    VforX = predict_V(Xt, A)
    y_predict= predict_y(VforX, a_0)
    error_in_y = normalized_residual(yt, y_predict)
    X_predict = np.matmul(VforX, A)
    col_wise_X_error = colwise_matrix_residual(Xt, X_predict)

    error_in_X= matrix_residual(Xt, X_predict)
    if returnA:
        return error_in_y, error_in_X, col_wise_X_error,y_predict, A, a_0
    else:
        return error_in_y, error_in_X, col_wise_X_error, y_predict


def predict_V(X, A):

    '''This function makes a prediction of the matrix V frm X_test and learnt coefficients A. Used when making a
    prediction for y_test.'''

    AAt = np.matmul(A, np.transpose(A))

    try:

        G = np.linalg.solve(AAt, A)
        v=np.matmul(X, np.transpose(G))
    except:
        AXt = np.matmul(A, np.transpose(X))
        print('########################## Had to use lstsq ########################')
        Vt=np.linalg.lstsq(AAt,AXt, rcond=None)[0]
        v = np.transpose(Vt)

    return v


def predict_y(v, a_0):
    '''Makes a prediction of y_test'''
    hat_y = np.matmul(v, a_0)
    return hat_y


def matrix_residual(Mtrue, Mp):
    '''Computes the residual of the matrix prediction. The order of the input parameters is not important. '''
    num=Mtrue-Mp
    return np.linalg.norm(num, 'fro')


def colwise_matrix_residual(Mtrue, Mp):
    '''Computes the normalized columnwise residual of the matrix prediction, the order of the input parameters is
    important. It is possible that some columns of the matrix have norm zero since some features only appear in some
    varieties.'''
    num=Mtrue-Mp
    den = np.sqrt(np.sum(Mtrue**2, axis=0))
    if np.any(den!=0):
        with np.errstate(divide='ignore', invalid='ignore'):
            return (np.sqrt(np.sum(num**2, axis=0)))/den
    else:
        return np.nan
