import scipy
import pandas as pd
from Regressors import normalized_residual, squared_res, unit_norm_zero_mean, predict_V, predict_y, matrix_residual, V, degree_spec_regression
from Plots import *
import warnings


def model_training_predicting(yv, X, y_test, X_test, rk, res_df, info,cviter):
    '''This function performs regression and chooses the parameter (number of segments) by cross validation.
    It takes the following input parameters:
    - yv: vector of measured y training values,
    - X: array of independent features for training,
    - y_test: vector of measured y test values,
    - X_test: array of independent features for testing,
    - rk: scalar, maximum number of segments to be considered during cross validation,
    - res_df: residual dataframe where to store the results of regression,
    - info: tuple of strings containing training and testing specifications,
    - cviter: list of cross validation splits'''

    #initialize the residual arrays, as many rows as cv splits, as many columns as possible numbers of segments (min 2)
    res_arrayy=np.zeros((len(cviter), len(range(2, rk+1))))
    res_arrayX=np.zeros((len(cviter), len(range(2, rk+1))))

    i=0
    # we loop over the cross validation splits and at every iteration we store the model metrics, residual in y and in X,
    # in the initialized arrays
    for vali, cvtrain in cviter:
        y_train = np.array(yv[cvtrain])
        X_train = np.array(X[cvtrain, :])
        y_vali = np.array(yv[vali])
        X_vali = np.array(X[vali, :])
        #for a given cv split we loop over the possible numbers of segments
        for deg in range(2, rk + 1):
            Vs ,W = V(y_train, deg)
            if isinstance(Vs, int):
                # for this number of segments the matrix V is singular
                res_arrayy[i, deg - 2] = np.nan
                res_arrayX[i, deg - 2] = np.nan
            else:
                # in this case the matrix V is not singular and we can go on
                error_in_yCV, error_in_XCV, colwise_err_X_CV, yhat=degree_spec_regression(Vs, W, y_train, X_train, y_vali, X_vali)
                if np.all(np.isnan(colwise_err_X_CV)):
                    print(f'{info} CV curve degree {deg} Xt has zero norm in the columns')
                res_arrayy[i, deg-2]=error_in_yCV
                res_arrayX[i, deg-2]=error_in_XCV
        i+=1

    # we take the average of the residual arrays over the cv splits so we can find the optimal number of segments
    warnings.filterwarnings(action='ignore', message='Mean of empty slice')
    res_arrayy=np.nanmean(res_arrayy, axis=0)
    res_arrayX=np.nanmean(res_arrayX, axis=0)

    # we store the metrics for each number of segments into the result dataframe
    for j in range(res_arrayy.shape[0]):
        if not np.isnan(res_arrayX[j].item()):
            new_index = res_df.shape[0] + 1

            res_df.loc[new_index, ['Model','Dataset','Trained on field','Trained on year',
                                   'Tested on field','Tested on year','Degree',
                                   'Residual X', 'Rank X']] =('CV',info[2], info[0][1], info[0][0], info[1][1], info[1][0],
                                                              range(2, rk + 1)[j], res_arrayX[j].item(), rk)
            res_df.loc[new_index, ['Residual y test']] = res_arrayy[j].item()

    # store the optimal number of segments in the variable degmin
    try:
        degmin=np.nanargmin(res_arrayX).item()+2
    except:
        degmin=2


    '''Test residual curve'''
    # We loop over the possible number of segments and compute the model performance on the test set
    for deg in range(2, rk+1):
        Vs, W = V(yv, deg)
        if not  isinstance(Vs, int):

            error_in_y, error_in_X, col_w_err_X, yhat=degree_spec_regression(Vs, W, yv, X, y_test, X_test)
            if  np.all(np.isnan(col_w_err_X)):
                print(f'{info} test curve degree {deg} Xt has zero norm in the columns')
            new_index = res_df.shape[0] + 1
            res_df.loc[new_index, ['Model','Dataset','Trained on field','Trained on year',
                                   'Tested on field','Tested on year','Degree',
                                   'Residual X', 'Rank X']] ='Test residual',f'{info[2]}', info[0][1], info[0][0], info[1][1], info[1][0], deg, error_in_X.item(), rk
            res_df.loc[new_index, ['Residual y test']] = error_in_y.item()

    '''Optimal degree in X '''

    # For the optimal number of segments we perform regression and measure the performance on the testset
    Vs, W = V(yv, degmin)
    ordering = np.argsort(np.asarray(res_arrayX))

    counter=0
    while isinstance(Vs, int):
        i, = np.where(ordering == counter)
        degmin=int(i[0] + 2)
        Vs, W = V(yv, int(degmin))
        counter+=1

    error_in_y, error_in_X, col_wise_X_t_residual,yhat,  AX, a_0x =degree_spec_regression(Vs, W, yv, X, y_test, X_test,True)
    if np.all(np.isnan(col_wise_X_t_residual)):
        print(f'{info} optimal degree in X Xt has zero norm in the columns')
    new_index = res_df.shape[0] + 1

    res_df.loc[new_index, ['Model','Dataset','Trained on field','Trained on year',
                        'Tested on field','Tested on year', 'Degree',
                           'Residual X', 'Rank X', 'P. Corr yt yt_hat', 'p-val']] = ('Optimal degree X',f'{info[2]}', info[0][1], \
        info[0][0], info[1][1], info[1][0],degmin, error_in_X.item(), rk, scipy.stats.pearsonr(y_test, yhat)[0],
                                                                                     scipy.stats.pearsonr(y_test, yhat)[1])
    res_df.loc[new_index, ['Residual y test']] = error_in_y.item()


    return res_df, int(degmin), a_0x, AX, col_wise_X_t_residual



def column_selection_based_on_X_res(colwise_X_res, y_train, y_test, X_train, X_test,  Areg, a0reg, residual_df, rk,
                                    info, optdeg, result_folder, ytrain_vb, ytest_vb):
    '''This function performs feature selection based on the provided column-wise residuals. It does not recompute
    coefficients, but chooses the optimal threshold based on the residual of y_training. It takes as input:
     - colwise_X_res: the array of column-wise residuals for the matrix X_test in the model with the optimal number of
                      segments,
     - y_train: the array of the training dependent variable y,
     - y_test: the array of the testing dependent variable y,
     - X_train: the matrix of independent variables for training,
     - X_test: the matrix of independent variables for testing,
     - Areg: the matrix of learned coefficients in the optimal model,
     - a0reg: the vector of learned coefficients in the optimal model,
     - residual_df: the dataframe where the results should be stored,
     - rk: the maximum number of segments allowed,
     - info: a string containing training and testing specifications,
     - optdeg: the optimal number of segments used in the model,
     - result_folder: path to the folder where results should be stored,
     - ytrain_vb: variety and batch labels for y_train,
     - ytest_vb: variety and batch labels for y_test.'''

    # We are careful to exclude potential non-numeric elements from the columnwise residuals array, then we sort these
    # residuals and store the filtered array in the variable thresholds
    colwise_X_res[~np.isfinite(colwise_X_res)] = np.nan
    thresholds =np.sort(colwise_X_res)[int(2*optdeg):]
    thresholds = thresholds[~np.isnan(thresholds)]

    # We initialise lists in which to store the metrics of interest
    res_X_test = []
    res_y_train = []
    res_y_test = []
    correlations=[]
    ytr_hats=[]
    pvalues=[]
    yhats=[]

    # we iterate over thresholds, each value yields a subset of the columns, we start with the smallest set
    for threshold in thresholds:
        filter = np.where(colwise_X_res <= threshold)[0]

        # we can simply apply the filter on the matrix A, there is no need to recompute
        X_test_filtered = X_test[:, filter]
        A=Areg[:,filter]
        X_filtered = X_train[:, filter]
        # in the following we compute all metrics of interest and store them in the previously initialized lists.
        # We also store the predictions made for y_test by every partition.
        v_tilde = predict_V(X_test_filtered, A)
        m=v_tilde.shape[1]-a0reg.shape[0]

        if m==0:
            pa0reg = a0reg
        else:
            pa0reg = np.pad(a0reg, (0, m), constant_values=0)

        y_predict = predict_y(v_tilde, pa0reg)
        res_X = matrix_residual(X_test_filtered, np.matmul(v_tilde, A))
        res_X_test.append(res_X.item())
        vfortr=predict_V(X_filtered, A)
        m=vfortr.shape[1]-a0reg.shape[0]
        if m==0:
            pa0reg = a0reg
        else:
            pa0reg = np.pad(a0reg, (0, m), constant_values=0)
        y_tr_pred = predict_y(vfortr, pa0reg)
        err_y_tr = normalized_residual(y_train, y_tr_pred)
        error_in_y = normalized_residual(y_test, y_predict)
        res_y_test.append(error_in_y.item())
        res_y_train.append(err_y_tr.item())
        try:
            corr, pval=scipy.stats.pearsonr(y_test, y_predict)
        except:

            corr=0
            pval=0
        correlations.append(corr)
        ytr_hats.append(y_tr_pred)
        pvalues.append(pval)
        yhats.append(y_predict)

    # the optimal threshold is the one for which the residual on y_training is minimized
    res_y_train=np.array(res_y_train)
    thropt = thresholds[np.nanargmin(res_y_train)]

    # we build a boolean filter based on which columns have a smaller residual than the chosen threshold
    BinFilter = colwise_X_res <= thropt

    # we add a row to the result dataframe for this model
    new_index = residual_df.shape[0] + 1
    residual_df.loc[new_index, ['Model','Dataset', 'Trained on field','Trained on year',
                        'Tested on field','Tested on year', 'Degree', 'Residual X'
        , 'Rank X']] ='Col. Sel.',f'{info[2]}', info[0][1], info[0][0], info[1][1], info[1][0], optdeg, res_X_test[np.argmin(res_y_train)], rk
    residual_df.loc[new_index, ['Residual y test',
                                '# of columns', 'Optimal Threshold','P. Corr yt yt_hat', 'p-val']] = res_y_test[np.argmin(res_y_train)], np.sum(BinFilter), \
        thropt,correlations[np.argmin(res_y_train)], pvalues[np.argmin(res_y_train)]


    unit_norm_y = unit_norm_zero_mean(y_test)
    unit_norm_yhat = unit_norm_zero_mean(yhats[np.argmin(res_y_train)])
    residual_unit = squared_res(unit_norm_y, unit_norm_yhat)
    corr = scipy.stats.pearsonr(unit_norm_y, unit_norm_yhat)
    residual_df.loc[new_index, ['Unit residual test', 'Pearson unit y yhat', 'unit p-val']] = residual_unit, corr[0], \
    corr[1]

    check1=len(ytest_vb)==len(y_test)
    if check1:

        testdf = pd.DataFrame({'Variety':ytest_vb['Variety'].values, 'Batch':ytest_vb['Batch'].values,'Y': y_test, 'Yhat': yhats[np.argmin(res_y_train)]})
        testdf['Train/Test'] = 'test'
        for var in ytest_vb['Variety'].unique():
            test_var = testdf[testdf['Variety'] == var]
            yvar = test_var['Y'].values
            yhatvar = test_var['Yhat'].values

            unit_norm_y = unit_norm_zero_mean(yvar)
            unit_norm_yhat = unit_norm_zero_mean(yhatvar)
            if len(unit_norm_y) > 2:
                # the pearson correlation coefficient of two vectors of length 2 is always statistically insignificant
                residual = squared_res(unit_norm_y, unit_norm_yhat)
                residual_df.loc[new_index, f'Unit residual test {var}'] = residual
                corr = scipy.stats.pearsonr(unit_norm_y, unit_norm_yhat)
                residual_df.loc[new_index, f'Pearson unit y yhat {var}'] = corr[0]
                residual_df.loc[new_index, f'unit p-val {var}'] = corr[1]
    # We build  dataframe to save the predictions made for both the training and testing sets by the reduced model
    traindf = pd.DataFrame({'Variety':ytrain_vb['Variety'].values, 'Batch':ytrain_vb['Batch'].values,'Y': y_train, 'Yhat': ytr_hats[np.argmin(res_y_train)]})
    traindf['Train/Test'] = 'train'
    pd.concat([traindf, testdf], ignore_index=True).to_csv(os.path.join(result_folder, f'{info}_col_sel_y_yhat.csv'), index=False)


    return residual_df, BinFilter, thropt


def experiment_pipeline(y_train, y_test, X_train, X_test, residual_df,cviter, train_info, test_info, tuber_data_info,
                        result_folder, maximum_basis_size, X_columns):
    '''This function performs model training, chooses with cross validatio (CV) the appropriate number of segments for
    the DG scheme, and for the optimal number of segments performs feature selection. The function returns a dataframe
    collecting all performance metrics and parameters for each run (all possible segment numbers evaluated on the
    validation set and on the test set plus one run of feature selection for the optimal number of segments.
    This function has input parameters:
    - y_train: dataframe containing three columns for training (Variety, Batch, measured_y) where the measured y values
               are in the last column,
    - y_test: dataframe containing three columns for testing (Variety, Batch, measured_y) where the measured y values
               are in the last column,
    - X_train: dataframe containing p+2 columns for training where p columns contain the seed tuber features and the
               first two columns contain the Variety and Batch labels,
    - X_test: dataframe containing p+2 columns for testing where p columns contain the seed tuber features and the
               first two columns contain the Variety and Batch labels,
    - residual_df: dataframe where the regression results should be stored. This can be empty, or if the function is
               called in a loop, it can be the result of a previous regression.
    - cviter: list of cross validation splits as returned from the function Main_regression_setting.CV_divide_vector
    - train_info: string containing the information regarding the field and year we are training on
    - test_info: string containing the information regarding the field and year we are testing on
    - tuber_data_info: string containing the information regarding the tuber data used in the model
    - result_folder: path where the results of regression should be stored
    - maximum_basis_size: maximum number of segments for the DG scheme, choice to be made by the user based on the size
                          of the dataset,
    - X_columns: names of the features used in the model.

    '''
    plt.close('all')

    # we separate training and testing dataframes into their numerical part and the labels
    y_train_var_batc=y_train.loc[:,['Variety','Batch']].copy()
    y_test_var_batch=y_test.loc[:,['Variety','Batch']].copy()
    y_train=np.array(y_train.iloc[:, -1])
    y_test=np.array(y_test.iloc[:, -1])
    X_train=np.array(X_train.iloc[:, 2:])
    X_test=np.array(X_test.iloc[:, 2:])

    # In the following function we run the CV procedure to find the optimal number of segments optdeg, obtain the optimal
    # coefficients Areg, a0reg, as well as the columnwise residuals for the predictions of X_test to be used in column
    # selection
    residual_df,optdeg, a0reg, Areg, colwise_X_res= (
    model_training_predicting( y_train, X_train,  y_test, X_test,  maximum_basis_size, residual_df,
                                  (train_info, test_info, tuber_data_info),cviter))

    # In the following we perform feature selection, we obtain a boolean filter of the same size as the columns of X,
    # indicating which columns have been rejected as well as the value of the threshold.
    info= (train_info, test_info, tuber_data_info)
    residual_df, BinFilt, optthresh=column_selection_based_on_X_res(colwise_X_res, y_train, y_test, X_train, X_test,
                                                                    Areg, a0reg, residual_df, maximum_basis_size, info,
                                                                    optdeg, result_folder, y_train_var_batc,
                                                                    y_test_var_batch)

    # We build and save a dataframe where we store for each column name whether it was kept and what residual it had
    (pd.DataFrame({'Column':X_columns, 'Kept':BinFilt, 'Residual':colwise_X_res}).
     to_csv(os.path.join(result_folder, f'{info}_col_sel_kept_columns.csv')))


    # we call a plotting function to visualize which columns have been rejected by the column selection procedure
    info= f'Col.wise-residuals {tuber_data_info}_trained_on_{train_info}_tested_on_{test_info}'
    plot_col_selection_residuals(colwise_X_res,  BinFilt,optthresh, info, result_folder, X_columns)

    return residual_df